/*
 *  Copyright (C) 2009 Mayibongwe S. Nkambule <msnkambule@gmail.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 */


#include "window.h"

#include <QtGui>
#include <KApplication>
#include <KFileDialog>
#include <KDialog>
#include <KMessageBox>
#include <KLocalizedString>
#include <KPushButton>
#include <KHelpMenu>
#include <KMenu>
#include <KDialogButtonBox>
#include <KGuiItem>
#include <KVBox>
#include <KProgressDialog>
#include <KConfigGroup>
#include <KSharedConfig>

#ifdef HAVE_TAGLIB
	#include <taglib/tag.h>
	#include <taglib/fileref.h>
	#include "taglib_filetype_resolvers/mimefiletyperesolver.h"
	#include "tageditor.h"	
	
#endif

enum Column
	{
		STATUS = 0, //shown at third column
		SOURCEFILE,
		OUTFILE,
		TITLE,
		ARTIST,
		ALBUM,
		GENRE,
		TRACK,
		YEAR,
		COMMENT
	};

// special strings sorter. For strings containing integers with a suffix
// eg. ["120 Hz", "890 Hz", "89 Hz" ] where Hz is suffix
QStringList mysort(QStringList &list, QString suffix)
{
	QList<int> nums;
	for(int i=0; i<list.size(); i++)
	{
		int v = list[i].split(" ")[0].toInt();
		nums << v;
	}
	
	qSort(nums.begin(), nums.end());
	list.clear();
	for (int i=0; i<nums.size(); i++)
		list << QString::number(nums[i]) + " " + suffix;
	return list;
}

// special strings sorter. For strings containing doubles with a suffix
// eg. ["120 Hz", "890 Hz", "89 Hz" ] where Hz is suffix
QStringList mydoublesort(QStringList &list, QString suffix)
{
	QList<double> nums;
	for(int i=0; i<list.size(); i++)
	{
		double v = list[i].split(" ")[0].toDouble();
		nums << v;
	}
	
	qSort(nums.begin(), nums.end());
	list.clear();
	for (int i=0; i<nums.size(); i++)
		list << QString::number(nums[i],'f', 2) + " " + suffix;
	return list;
}

CWindow::CWindow(QWidget *parent, Qt::WindowFlags f )
   : KDialog(parent,f)
{
    setWindowTitle("Konvertible");
	readSettings();

#ifdef HAVE_TAGLIB
	m_model = new QStandardItemModel(0,10);
#else
	m_model = new QStandardItemModel(0,3);
#endif

	m_model->setHeaderData(STATUS, Qt::Horizontal, QObject::tr(""));
    m_model->setHeaderData(SOURCEFILE, Qt::Horizontal, QObject::tr("Source File"));
    m_model->setHeaderData(OUTFILE, Qt::Horizontal, QObject::tr("Output File"));
	
#ifdef HAVE_TAGLIB
	TagLib::FileRef::addFileTypeResolver(new  MimeFileTypeResolver );
	
	m_model->setHeaderData(TITLE, Qt::Horizontal, QObject::tr("Title"));
	m_model->setHeaderData(ARTIST, Qt::Horizontal, QObject::tr("Artist"));
	m_model->setHeaderData(ALBUM, Qt::Horizontal, QObject::tr("Album")); 
	m_model->setHeaderData(GENRE, Qt::Horizontal, QObject::tr("Genre")); 
	m_model->setHeaderData(TRACK, Qt::Horizontal, QObject::tr("Track"));
	m_model->setHeaderData(YEAR, Qt::Horizontal, QObject::tr("Year"));
	m_model->setHeaderData(COMMENT, Qt::Horizontal, QObject::tr("Comment"));
#endif
	
	//
	m_filelistview = new QTreeView;
	m_filelistview->setWhatsThis("List of files to convert.\nRight click to"
									" show delete, rename, make FAT safe names options");
    m_filelistview->setAlternatingRowColors(true);
	m_filelistview->setRootIsDecorated(false);
    m_filelistview->setModel(m_model);
	m_filelistview->setSelectionMode(QAbstractItemView::ExtendedSelection);
	hideTagColumns();
	m_filelistview->resizeColumnToContents(STATUS);
	
	connect( m_filelistview, SIGNAL(pressed(QModelIndex)), this, SLOT(showMenu(QModelIndex))  );	

    addFileBtn = new QPushButton;
    addFileBtn->setText("&Add Files");
	addFileBtn->setWhatsThis("Select files you want to convert.");
    connect(addFileBtn, SIGNAL(clicked()), this, SLOT(addFiles()));

    QLabel *alabel = new QLabel("Destination:");
    outfolderEdit = new QLineEdit;
	outfolderEdit->setWhatsThis("The folder where output files will be saved in.");
    openfolderBtn = new KPushButton( KIcon("folder"), "");
	openfolderBtn->setWhatsThis("Click to choose destination folder.");
    connect(openfolderBtn, SIGNAL(clicked()), this, SLOT(setOutFolder()));
    QHBoxLayout * outLayout = new QHBoxLayout;
    outLayout->addWidget(alabel);
    outLayout->addWidget(outfolderEdit);
    outLayout->addWidget(openfolderBtn);

    QHBoxLayout *filelayout = new QHBoxLayout;
    filelayout->addWidget(addFileBtn);
    filelayout->addLayout(outLayout);

    //codec
    QLabel *clabel = new QLabel("Codec:");
    clabel->setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
    m_codecBox = new QComboBox;
	m_codecBox->setWhatsThis("Audio codec determines type of new files.");
    findCodecs();
    connect(m_codecBox, SIGNAL(currentIndexChanged(int)), SLOT(changeExtension()));
	//connect(m_codecBox, SIGNAL(currentIndexChanged(int)), SLOT(clearStatusIcons()));

    QHBoxLayout * cLayout = new QHBoxLayout;
    cLayout->addWidget(clabel);
    cLayout->addWidget(m_codecBox);
	
    //extension
    m_extension = new QLineEdit;
	m_extension->setWhatsThis("Extension of output files.\nType it if its blank.");
    connect( m_extension, SIGNAL(textChanged(QString)), this, SLOT(fixOutputFiles()) );
    QLabel *extlabel = new QLabel("Extension:");
    extlabel->setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
    QHBoxLayout * extLayout = new QHBoxLayout;
    extLayout->addWidget(extlabel);
    extLayout->addWidget(m_extension);
    //channels
    m_channelsBox = new QSpinBox;
	m_channelsBox->setWhatsThis("Audio channels.\n1 - Mono\n2 - Stereo");
    m_channelsBox->setMinimum(1);
    m_channelsBox->setValue(2);
	connect(m_channelsBox, SIGNAL(valueChanged(int)), SLOT(clearStatusIcons()));
    QLabel *chlabel = new QLabel("Channels:");
    chlabel->setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
    QHBoxLayout * chLayout = new QHBoxLayout;
    chLayout->addWidget(chlabel);
    chLayout->addWidget(m_channelsBox);
	
    //frequency
    m_freqBox = new QComboBox;
	m_freqBox->setWhatsThis("Low Frequency means low quality and small file size.");
	m_freqBox->addItems( QStringList() << frequencies << "Other" );
	m_curFreq = "44100 Hz";
	m_freqBox->setCurrentIndex( frequencies.indexOf(m_curFreq) );
	connect(m_freqBox, SIGNAL(activated(QString)), this, SLOT(freqChanged(QString)));
	connect(m_freqBox, SIGNAL(activated(int)), SLOT(clearStatusIcons()));	
    QLabel *freqlabel = new QLabel("Frequency:");
    freqlabel->setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
    QHBoxLayout * freqLayout = new QHBoxLayout;
    freqLayout->addWidget(freqlabel);
    freqLayout->addWidget(m_freqBox);
	
    //bitrate
    m_bitrBox = new QComboBox;
	m_bitrBox->setWhatsThis("Low bitrate means low quality and small file size.");
	m_bitrBox->addItems( QStringList() << bitrates << "Other" );
	m_curBitRate = "96.00 kbits/s";
	m_bitrBox->setCurrentIndex( bitrates.indexOf(m_curBitRate)  );
	connect(m_bitrBox, SIGNAL(activated(QString)), this, SLOT(bitRChanged(QString)));
	connect(m_bitrBox, SIGNAL(activated(int)), SLOT(clearStatusIcons()));	
    QLabel *bitlabel = new QLabel("Bitrate:");
    bitlabel->setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
    QHBoxLayout * bitLayout = new QHBoxLayout;
    bitLayout->addWidget(bitlabel);
    bitLayout->addWidget(m_bitrBox);   


    QGridLayout *optionsLayout = new QGridLayout;
    optionsLayout->addLayout(cLayout, 0, 0, 1, 2);
    optionsLayout->addLayout(extLayout, 0, 2);
    optionsLayout->addLayout(chLayout, 1, 0);
    optionsLayout->addLayout(freqLayout, 1, 1);
    optionsLayout->addLayout(bitLayout, 1, 2);

    m_convertButton = new QPushButton;
    m_convertButton->setText("&Convert");
    connect(m_convertButton, SIGNAL(clicked()), this, SLOT(convert()));

		
	QWidget *holder = new QWidget;;
	setMainWidget(holder);
    QVBoxLayout *layout = new QVBoxLayout(holder);
    layout->addLayout(filelayout);
    layout->addLayout(optionsLayout);
    layout->addWidget(m_convertButton);
    layout->addWidget(m_filelistview);
	
	showButtonSeparator( true );
	setButtons(Help | User1);
	setButtonGuiItem( User1, KStandardGuiItem::quit() );	
	KHelpMenu *helpMenu = new KHelpMenu(this, KGlobal::mainComponent().aboutData(), true);
    setButtonMenu( Help, helpMenu->menu() );
	connect( this, SIGNAL( user1Clicked() ), SLOT( close() ) );
	
    lastDestDir = QDir::currentPath();
    lastSourceDir = lastDestDir;

    changeExtension();	
	createActions();
	holder ->setFocus();
	
}

void CWindow::hideTagColumns()
{
	//m_filelistview->setColumnHidden(TITLE , true);
	//m_filelistview->setColumnHidden(ARTIST, true);
	//m_filelistview->setColumnHidden(ALBUM, true);
	m_filelistview->setColumnHidden(GENRE, true);
	m_filelistview->setColumnHidden(TRACK , true);
	m_filelistview->setColumnHidden(YEAR , true);
	m_filelistview->setColumnHidden(COMMENT, true);
}

// runs ffmpeg to get list of audio encoders. We can't continue if
// we can't find it.
void CWindow::findCodecs()
{
    QProcess ffmpeg;
    ffmpeg.start("ffmpeg", QStringList() << "-codecs"); //for newer ffmpeg. Fails in older versions.
    if (!ffmpeg.waitForStarted())
	{
		KMessageBox::error(this, 
			"<b>ffmpeg</b> could not be found. We use it to convert files."
			"<p>Install it using your package manager or"
			" download it from <a href='http://ffmpeg.org'>http://ffmpeg.org</a></p>",
			"",
			KMessageBox::AllowLink);
		exit(1);
	}


    if (!ffmpeg.waitForFinished())
	{
        KMessageBox::error(this, 
			"<b>ffmpeg</b> could not be found. We use it to convert files."
			"<p>Install it using your package manager or"
			" download it from <a href='http://ffmpeg.org'>http://ffmpeg.org</a></p>",
			"",
			KMessageBox::AllowLink);
		exit(1);
	}
	
	if ( ffmpeg.exitCode() != 0 ) //old ffmpeg
	{
		ffmpeg.start("ffmpeg", QStringList() << "-formats");
		ffmpeg.waitForStarted();
		ffmpeg.waitForFinished();
	}
    
    //  ffmpeg outputs lines like
    //      DEA     CODECNAME   CODECNAMEFULLNAME
    //      D A     CODECNAME   CODECNAMEFULLNAME
    //  We only want codecs which are for audio and can be encoded
    //  that is, first part should contain E and A
    QString result(ffmpeg.readAll());
    QStringList ffresultLs = result.split("\n");

    for (int i=0; i<ffresultLs.size(); i++)
    {
        QString line = ffresultLs[i].simplified();
        QStringList codecls = line.split(" ");
        if (codecls[0].contains("EA"))
        {
            codecls.removeAt(0);
            QString codec =  codecls.takeAt(0);
            m_codecList << codec;
            QString codecName = codecls.join(" ");
            
            m_codecBox->addItem(codecName + " (" +codec+")");
        }
    }
	
	if (m_codecList.size() == 0)
	{
		KMessageBox::sorry(this, 
			"No audio encoders found. Konvertible will not work."
			"<p>Your ffmpeg was found but it reports that it cannot encode audio."
			" Konvertible executed<br/> ffmpeg -formats<br/> and did not get a list of audio encoders.</p>",
			"",
			KMessageBox::AllowLink);
	}
}

void CWindow::changeExtension()
{
    if (m_codecList.size() == 0)
		return;
		
	int codecIndex = m_codecBox->currentIndex();
    
    QString codec = m_codecList[codecIndex];
    QString ext;
    
    if (codec == "aac" || codec =="libfaac")
        ext="aac";
    else if (codec == "ac3")
        ext="ac3";
    else if (codec == "libmp3lame")
        ext="mp3";
    else if (codec == "mp2")
        ext="mp2";
    else if (codec == "vorbis" || codec == "libvorbis")
        ext="ogg";
    else if (codec == "flac")
        ext="flac";
     else if (codec == "wmav1" || codec == "wmav2")
        ext="wma";
	else if (codec == "adpcm_ima_qt")
        ext="mov";
    else if ( codec.startsWith("pcm") || codec.startsWith("adpcm_") )
        ext="wav";
    else if (codec == "libopencore_amrnb")
        ext="amr";
	else if (codec == "roq_dpcm")
        ext="roq";	
	
    m_extension->setText(ext);
    fixOutputFiles();
}

void CWindow::fixOutputFiles()
{
	clearStatusIcons();

	QString newExt = m_extension->text();

	for (int i=0; i<m_model->rowCount(); i++)
	{
		QString oldName = m_model->item(i,OUTFILE)->text();
		QString outf = QFileInfo(oldName).completeBaseName() + "." + newExt;
		m_model->setData(m_model->index(i, OUTFILE), outf);
	}

	m_filelistview->resizeColumnToContents(OUTFILE);
}

// Adds and displays files in the conversion list
void CWindow::setFiles(QStringList files)
{
	//get list of already added files
	QStringList addedFiles;
	for (int i=0; i < m_model->rowCount(); i++)
		addedFiles << m_model->item(i,SOURCEFILE)->text();
	
	QString ext = m_extension->text();

	for (int i=0; i < files.size(); i++)
	{
		//don't add if already added
		if ( addedFiles.contains( files[i] ) )
			continue;
		else
			addedFiles << files[i];
			
		int r = m_model->rowCount();
		m_model->insertRow(r);
		m_model->setData(m_model->index(r, SOURCEFILE), files[i]);
		m_model->item(r, SOURCEFILE)->setFlags(Qt::ItemIsSelectable | Qt::ItemIsEnabled);
		//output name
		QString outf = QFileInfo(files[i]).completeBaseName() + "." + ext;
		m_model->setData(m_model->index(r, OUTFILE), outf);
		//meta data
#ifdef HAVE_TAGLIB
		TagLib::FileRef f( files[i].toLatin1() );
		if ( !f.isNull() && f.tag() )
		{
			TagLib::Tag *tag = f.tag();
			m_model->setData( m_model->index(r, TITLE), tag->title().toCString(true) );
			m_model->setData( m_model->index(r, ARTIST), tag->artist().toCString(true) );
			m_model->setData( m_model->index(r, ALBUM), tag->album().toCString(true) );
			//m_model->setData( m_model->index(r, COMPOSER), tag->composer().toCString(true) );
			m_model->setData( m_model->index(r, TRACK), QString::number(tag->track()) );
			m_model->setData( m_model->index(r, GENRE), tag->genre().toCString(true) );
			m_model->setData( m_model->index(r, YEAR), QString::number(tag->year()) );
			m_model->setData( m_model->index(r, COMMENT), tag->comment().toCString(true) );
		}
#endif
	}   
	m_filelistview->resizeColumnToContents(SOURCEFILE);
	m_filelistview->resizeColumnToContents(OUTFILE);	
}

// Called when user clicks Add Files button
void CWindow::addFiles()
{
	QStringList files = KFileDialog::getOpenFileNames
	(
		KUrl(lastSourceDir),
		"*|All Files\n*.flac|FLAC\n*.ogg|OGG\n*.mp3|MP3\n*.wav|Wave",
		this,
		"Add files to convert"
	);

	if (files.size() > 0)
	{
		lastSourceDir = QFileInfo(files[0]).absoluteDir().absolutePath() ;
		setFiles(files);
		clearStatusIcons();
	}   
}

// Called when user clicks the folder button
void CWindow::setOutFolder()
{
	QString folder = KFileDialog::getExistingDirectory
	(
		lastDestDir,
		this,
		"Choose destination folder"
	);
   
	if (folder.length() > 0)
	{
		lastDestDir = QFileInfo(folder).absoluteDir().absolutePath() ;      
		outfolderEdit->setText(folder);
		clearStatusIcons();
	}
}

void CWindow::createActions()
{
	removeAct = new QAction( KIcon("list-remove"), "Remove", this );
	connect(removeAct, SIGNAL(triggered()), this, SLOT(remove()));
	
	renameAct = new QAction( KIcon("edit-rename"), "Rename", this );
	connect(renameAct, SIGNAL(triggered()), this, SLOT(rename()));
	
	toFATAct = new QAction(this);
	toFATAct->setText("Conform Name to FAT");
	connect(toFATAct, SIGNAL(triggered()), this, SLOT(makeFATName()));
	
	#ifdef HAVE_TAGLIB
	metaDataAct = new QAction(this);
	metaDataAct->setText("");
	//removeAct->setIcon(  );
	connect(metaDataAct, SIGNAL(triggered()), this, SLOT(editMetadata()));
	#endif
	
}


void CWindow::showMenu(QModelIndex index)
{
	//right click?
	if (QApplication::mouseButtons() != Qt::RightButton)
		return;
	
	QMenu menu(this);
	
	if ( index.column() == OUTFILE )
	{
		menu.addAction(renameAct);
		menu.addAction(toFATAct);
		menu.addSeparator();
	}
	#ifdef HAVE_TAGLIB
		metaDataAct->setText("Edit Track Details");
		menu.addAction(metaDataAct);
	#endif
	
	menu.addSeparator();
	menu.addAction(removeAct);	
	menu.exec(QCursor::pos());
}

void CWindow::rename()
{
	m_filelistview->edit( m_filelistview->currentIndex() );
	clearStatusIcons();
	m_filelistview->resizeColumnToContents(OUTFILE);
}


// removes selected files
void CWindow::remove()
{
	QItemSelectionModel   *selection = m_filelistview->selectionModel();
	QList<QPersistentModelIndex> indexLs;	
	
	//get persistent indexes, otherwise we may remove wrong files
	for(int i=0; i<selection->selectedIndexes().size(); i++)
	{
		QPersistentModelIndex id(selection-> selectedIndexes()[i]);
		indexLs.append(id);
	}
	
	for(int i=0; i<indexLs.size(); i++)
		m_model->removeRow( indexLs[i].row() );
		
	clearStatusIcons();
}

// remove |\?*<":>+[]/ in filenames
// Necessary when files are to be transferred to VFat filesystems like
// memory cards, flash disks, etc
void CWindow::makeFATName()
{
	QItemSelectionModel   *selection = m_filelistview->selectionModel();
	QList<QPersistentModelIndex> indexLs;	
	
	//get persistent indexes
	for (int i=0; i<selection->selectedIndexes().size(); i++)
	{
		QPersistentModelIndex id(selection-> selectedIndexes()[i]);
		indexLs.append(id);
	}
	
	for (int i=0; i<indexLs.size(); i++)
	{
		QString oldName = m_model->item(indexLs[i].row(),OUTFILE)->text();
		for (int c=0; c<oldName.length(); c++)
		{
			QChar ch = oldName[c];
			if (ch=='|' ||  ch=='\\' || ch=='?' || ch=='*' || ch=='<' || ch=='"' || 
					ch==':' || ch=='>' || ch=='+' || ch=='[' || ch==']' || ch=='/' )
			oldName.remove(c, OUTFILE);
		}
		m_model->setData(m_model->index(indexLs[i].row(), OUTFILE), oldName);
	}
	clearStatusIcons();
}

void CWindow::editMetadata()
{
	#ifdef HAVE_TAGLIB
	
	QStringList tagList;
	
	int r = m_filelistview->currentIndex().row();
	tagList << m_model->item(r,TITLE)->text()
					<< m_model->item(r,ARTIST)->text()
					<< m_model->item(r,ALBUM)->text()
					<< m_model->item(r,GENRE)->text()
					<< m_model->item(r,TRACK)->text()
					<< m_model->item(r,YEAR)->text()
					<< m_model->item(r,COMMENT)->text();
	
	TagEditor *editor = new TagEditor;
	editor->setTags( tagList );
	int result = editor->exec();
	if ( result == QDialog::Accepted )
	{
		tagList = editor->tags();
		m_model->setData( m_model->index(r, TITLE), tagList[0] );
		m_model->setData( m_model->index(r, ARTIST), tagList[1] );
		m_model->setData( m_model->index(r, ALBUM), tagList[2] );
		m_model->setData( m_model->index(r, TRACK), tagList[3] );
		m_model->setData( m_model->index(r, GENRE), tagList[4] );
		m_model->setData( m_model->index(r, YEAR), tagList[5] );
		m_model->setData( m_model->index(r, COMMENT), tagList[6] );
	}
	
#endif
}


void CWindow::setProgressDialog()
{
    m_progressDlg = new KProgressDialog(this);
	m_progressDlg->setCaption("Converting");   
	QProgressBar *progress = m_progressDlg->progressBar();
	progress->setMaximum(0);
	connect( m_progressDlg, SIGNAL(cancelClicked()), this, SLOT(cancel()) );
}

void CWindow::updateMessage()
{
    if ( m_ffmpegProc->state() == QProcess::Running )
    {
		QString curFile = QFileInfo( m_model->item(m_converted,SOURCEFILE)->text() ).fileName();
		int left = m_model->rowCount() - (m_converted+1);
        m_progressDlg->setLabelText("<b>Converting " + curFile +"</b>"
					"<br/>" + QString::number(left) +" " +  (left==1?" file":"files") +" left.");
    }
}


void CWindow::ffmpegDone(int exitCode, QProcess::ExitStatus exitStatus)
{        
	if (exitCode != 0)
	{
		//remove 0 byte file ffmpeg may have created
		QFile::remove( outfolderEdit->text()+"/"+ m_model->item(m_converted,OUTFILE)->text() );
		
		//show fail icon
		KIcon icon("konvertible_fail");
		QStandardItem *iconItem = new QStandardItem(icon, "");
		iconItem->setFlags( Qt::ItemIsEnabled );
		m_model->setItem(m_converted, STATUS, iconItem);
		
		m_errors++;
		m_errMsg += "<br><br> ***" + m_model->item(m_converted,SOURCEFILE)->text()
					+ "***<br>"	+ m_ffmpegProc->readAll();			
	}
	else
	{
		//write tags
#ifdef HAVE_TAGLIB		
		TagLib::FileRef f( (outfolderEdit->text()+"/"+ m_model->item(m_converted,OUTFILE)->text()).toLatin1() );
		if (!f.isNull())
		{
			TagLib::Tag *tag = f.tag();
			tag->setTitle( TagLib::String(m_model->item(m_converted,TITLE)->text().toUtf8().data(), TagLib::String::UTF8 ) );
			tag->setArtist( TagLib::String(m_model->item(m_converted,ARTIST)->text().toUtf8().data(), TagLib::String::UTF8 ) );
			tag->setAlbum( TagLib::String(m_model->item(m_converted,ALBUM)->text().toUtf8().data(), TagLib::String::UTF8 ) );
			tag->setYear( m_model->item(m_converted,YEAR)->text().toInt() );
			tag->setTrack( m_model->item(m_converted,TRACK)->text().toInt() );
			tag->setGenre( TagLib::String(m_model->item(m_converted,GENRE)->text().toUtf8().data(), TagLib::String::UTF8 ) );
			tag->setComment( TagLib::String(m_model->item(m_converted,COMMENT)->text().toUtf8().data(), TagLib::String::UTF8 ) );
			f.save();
		}
#endif		
		//show success icon
		KIcon icon("konvertible_success");
		QStandardItem *iconItem = new QStandardItem(icon, "");
		iconItem->setFlags( Qt::ItemIsEnabled );
		m_model->setItem(m_converted, STATUS, iconItem);
	}

	m_converted++;	
			
	//convert next file, otherwise show message to user 
	if ( m_converted < m_model->rowCount() )
	{
		m_ffmpegProc->close();
		//delete m_ffmpegProc;
		runffmpeg(); //next file
	}
	else //we are done
	{
		m_progressDlg->close(); //close it and use KDE's message boxes now		
		if (m_errors > 0)
		{
			QString msg;
			if (m_errors < m_model->rowCount())
				msg = QString::number(m_errors)+ (m_errors ==1?" file":" files") + "  could not be converted.";
			else
				msg = "Sorry. Could not convert.\nSee details.";
			
			KMessageBox::detailedSorry(this, msg, m_errMsg, "");
		}
		else
			KMessageBox::information(this,	"Converted", "Done");	            
	}
}

// Called when user clicks the Cancel button in the progress dialog
void CWindow::cancel()
{
	m_ffmpegProc->disconnect();
	m_ffmpegProc->kill();
	m_progressDlg->hide();	
}


void CWindow::runffmpeg()
{
    QString outfile = outfolderEdit->text()+"/"+ m_model->item(m_converted,OUTFILE)->text();
	QString infile = m_model->item(m_converted,SOURCEFILE)->text();
	QString freq = m_freqBox->currentText().section(' ', 0,0);
	QString bitrate = m_bitrBox->currentText().section(' ', 0,0);
	
	QStringList args;
    args << "-i" << infile
            << "-acodec" << m_codecList[ m_codecBox->currentIndex() ]
            << "-ar" << freq
            << "-ac" << m_channelsBox->cleanText()
            << "-ab" << bitrate +"k"
            << "-y"
            << outfile;
			//<< "-map_meta_data"
			//<< outf +":"+inf;
    
    m_ffmpegProc = new QProcess(this);
	m_ffmpegProc->setProcessChannelMode(QProcess::MergedChannels);
    connect( m_ffmpegProc, SIGNAL(started()), this, SLOT(updateMessage()) );
    connect( m_ffmpegProc, SIGNAL(finished(int,QProcess::ExitStatus)), this,
                    SLOT(ffmpegDone(int, QProcess::ExitStatus)) );
                    
    m_ffmpegProc->start("ffmpeg", args);
}

void CWindow::convert()
{
	if (m_codecList.size() == 0)
	{
		KMessageBox::sorry(this, 
			"No audio encoders found. Konvertible will not work."
			"<p>Your ffmpeg was found but it reports that it cannot encode audio."
			" Konvertible executed<br/> ffmpeg -formats<br/> and did not get a list of audio encoders.</p>",
			"",
			KMessageBox::AllowLink);
			return;
	}
	
	if ( m_model->rowCount() == 0)
    {
        addFiles();
        return;
    }
        
    QString destfolder = outfolderEdit->text();
    if (destfolder.isEmpty() || !QFileInfo(destfolder).exists())
    {
        setOutFolder();
        return;
    }

    clearStatusIcons();
    setProgressDialog();
    m_converted = 0;
    m_errors = 0;
	m_errMsg.clear();
    runffmpeg();   
    m_progressDlg->show();   
}

// A slot called when user makes a selection in the frequency combo box.
// It traps selection of Other so that we show an input dialog for user to
// own frequency.
void CWindow::freqChanged(QString freq)
{
	if ( freq != "Other" )
	{
		m_curFreq = freq;
		return;
	}
	
	//ask user to enter own choice
	QSpinBox *freqEdit = new QSpinBox;
	freqEdit->setRange(0, 9999999);
	freqEdit->setValue( m_curFreq.split(" ")[0].toInt() );
	freqEdit->setSuffix(" Hz");
	
	KDialog dialog;
	dialog.setCaption("Frequency");
	dialog.setMainWidget(freqEdit);
	
	if (dialog.exec() == QDialog::Accepted)
	{
		m_curFreq = freqEdit->text();
		frequencies << m_curFreq;
		frequencies.removeDuplicates();
		mysort( frequencies, "Hz" );
		m_freqBox->clear();
		m_freqBox->addItems( QStringList() << frequencies << "Other" );
		m_freqBox->setCurrentIndex( frequencies.indexOf(m_curFreq) );
	}
	else //do this to prevent showing of Other if user cancelled
		m_freqBox->setCurrentIndex( frequencies.indexOf(m_curFreq) );
}


// A slot called when user makes a selection in the bitrate combo box.
// It traps selection of Other so that we show an input dialog for user to
// own bitrate.
void CWindow::bitRChanged(QString bitR)
{
	if( bitR != "Other" )
	{
		m_curBitRate = bitR;
		return;
	}
	
	//ask user to enter own choice
	QDoubleSpinBox *bitEdit = new QDoubleSpinBox;
	bitEdit->setRange(0, 9999999);
	bitEdit->setValue( m_curBitRate.split(" ")[0].toDouble() );
	bitEdit->setSuffix(" kbits/s");
	
	KDialog dialog;
	dialog.setCaption("Bitrate");
	dialog.setMainWidget(bitEdit);
	
	if (dialog.exec() == QDialog::Accepted)
	{
		m_curBitRate = QString::number( bitEdit->value(), 'f', 2 )  + " kbits/s";
		bitrates << m_curBitRate;
		bitrates.removeDuplicates();
		mydoublesort( bitrates, "kbits/s" );
		m_bitrBox->clear();
		m_bitrBox->addItems( QStringList() << bitrates << "Other" );
		m_bitrBox->setCurrentIndex( bitrates.indexOf(m_curBitRate) );
	}
	else //do this to prevent showing of Other if user cancelled
		m_bitrBox->setCurrentIndex( bitrates.indexOf(m_curBitRate) );
		
	clearStatusIcons();
}

// remove fail/success icons when user changes something
void CWindow::clearStatusIcons()
{
	for (int i=0; i <m_model->rowCount(); i++)
	{
		QStandardItem *iconItem = new QStandardItem( QIcon(), "" );
		iconItem->setFlags( Qt::ItemIsEnabled );
		m_model->setItem( i, STATUS, iconItem );
	}
}

void  CWindow::keyPressEvent( QKeyEvent * event )
{
	if ( event->key() == Qt::Key_Delete )
	{
		remove();
		event->accept();
	}	
}

void CWindow::closeEvent( QCloseEvent * event )
{
	writeSettings();
	event->accept();
}

void CWindow::readSettings()
{
	frequencies << "8000 Hz"<< "16000 Hz" << "22050 Hz" << "44100 Hz"
						<< "48000 Hz" << "96000 Hz";
						
	bitrates <<  "12.20 kbits/s"<<"64.00 kbits/s"<< "80.00 kbits/s" << "96.00 kbits/s" << "128.00 kbits/s"
					<< "160.00 kbits/s" << "192.00 kbits/s" << "224.00 kbits/s";
					
	KSharedConfigPtr config = KSharedConfig::openConfig();
	KConfigGroup generalGroup(config, "General" );
	
	frequencies = generalGroup.readEntry( "Frequencies", frequencies);
	bitrates = generalGroup.readEntry( "Bitrates", bitrates);
	
	resize( generalGroup.readEntry( "Size", QSize(780,377) ) );
}

void CWindow::writeSettings()
{
	KSharedConfigPtr config = KSharedConfig::openConfig();
	KConfigGroup generalGroup(config, "General" );
	
	generalGroup.writeEntry( "Frequencies", frequencies);
	generalGroup.writeEntry( "Bitrates", bitrates);
	generalGroup.writeEntry( "Size", size() );
	
	generalGroup.config()->sync();
}

/******************************   TAG Dialog ********************************************************/
